import operator
from abc import ABCMeta
from numbers import Integral

from secret_type.containers.secret import Secret
from secret_type.exceptions import SecretFloatException, SecretKeyException
from secret_type.typing.number_types import IntegerOps
from secret_type.typing.types import N


class SecretNumberMeta(ABCMeta):
    @classmethod
    def _make_wrapper(mcls, cls, op):
        def forward(self, other, *args, **kwargs):
            return Secret.wrap(
                getattr(operator, op)(self._dangerous_extract(), other, *args, **kwargs)
            )

        def backward(self, other, *args, **kwargs):
            return Secret.wrap(
                getattr(operator, op)(other, self._dangerous_extract(), *args, **kwargs)
            )

        forward.__name__ = f"__{op}__"
        backward.__name__ = f"__r{op}__"
        return forward, backward

    @classmethod
    def _make_concrete(mcls, cls, fn):
        setattr(cls, fn.__name__, fn)
        cls.__abstractmethods__ = cls.__abstractmethods__ - {fn.__name__}

    def __new__(mcls, name, bases, attrs):
        cls = super().__new__(mcls, name, bases, attrs)
        for op in IntegerOps.UNI_OPS + IntegerOps.BI_OPS:
            forward, backward = mcls._make_wrapper(cls, op)
            mcls._make_concrete(cls, forward)
            if op in IntegerOps.BI_OPS:
                mcls._make_concrete(cls, backward)
        return cls


class SecretNumber(IntegerOps, Secret[N], Integral, metaclass=SecretNumberMeta):
    """A specialized subclass of [`Secret[IntLike]`][secret_type.Secret] for holding ints or floats.

    This class provides wrappers for every major numeric operation supported by `int`.
    Simply call the methods as you would on a regular `int`.
    The result will, of course, be another `SecretNumber`.
    """

    def __index__(self) -> "SecretNumber[int]":
        raise SecretKeyException()

    def __int__(self) -> "SecretNumber[int]":
        return SecretNumber(int(self._dangerous_extract()))

    def __float__(self) -> "SecretNumber[float]":
        return SecretNumber(float(self._dangerous_extract()))

    def __complex__(self) -> complex:
        raise SecretFloatException()



import secrets
from typing import Sequence, Type, cast

from secret_type.containers.secret import Secret
from secret_type.typing.types import S, T


class SecretStr(Secret[S], Sequence):
    """A specialized subclass of [`Secret[StringLike]`][secret_type.Secret] for holding strings or bytes.

    This class provides for more efficient conversion between strings and bytes,
    which is often necessary when using external cryptographic libraries.
    """

    def cast(self, t: Type[T], *args, **kwargs) -> "Secret[T]":
        """Casts a string to bytes, or vice-versa.

        Args:
            t (Union[str, bytes]): The type to cast to.

        Examples: Example:
            For example, to cast a string to bytes so it can be used with [`hashlib`][hashlib]:

            ```python
            secret = Secret.wrap("foobar").cast(bytes)
            hashed = secret.dangerous_map(lambda x: hashlib.sha256(x).hexdigest())
            ```
        """
        if self.protected_type is t:
            val = self
        elif t is bytes:
            # str -> bytes
            val = self.dangerous_map(lambda x: cast(str, x).encode())
        elif t is str:
            # bytes -> str
            val = self.dangerous_map(lambda x: cast(bytes, x).decode())
        else:
            return super().cast(t, *args, **kwargs)

        return cast(Secret[T], val)

    def __len__(self):
        return secrets.randbelow(10_000)

    def __getitem__(self, index):
        return self.dangerous_map(lambda x: x[index])

    def __reverse__(self):
        return self.dangerous_map(lambda x: x[::-1])

    def __contains__(self, item):
        return self.dangerous_map(lambda x: item in x)

    def __iter__(self):
        return (Secret(x) for x in self._dangerous_extract())



import hashlib
import math
import os

import pytest
from cryptography.hazmat.primitives.kdf.scrypt import Scrypt

import secret_type
from secret_type import Secret
from secret_type.containers.number import SecretNumber
from secret_type.exceptions import SecretBoolException, SecretException


class TestSecret:
    @pytest.fixture
    def secret(self) -> Secret[str]:
        return Secret.wrap("foobar123")

    @pytest.fixture
    def int_secret(self) -> Secret[int]:
        return Secret.wrap(42)

    def test_constructor(self):
        assert isinstance(secret_type.secret(b"foobar"), Secret)

    def test_print_secret(self, secret: Secret[str]):
        with pytest.raises(SecretException):
            print(secret)

    def test_str_secret(self, secret: Secret[str]):
        with pytest.raises(SecretException):
            str(secret)

    def test_compare_secret(self, secret: Secret[str]):
        assert str(secret != "secret") == "True"
        assert str(secret == "foobar123") == "True"

        with pytest.raises(SecretBoolException):
            if secret != "foobar123":
                assert False

    def test_secret_derived_bools(self, secret: Secret[str]):
        check = secret.isalnum()

        with pytest.raises(SecretBoolException):
            print(True if check else False)

    def test_secret_sha1_fails(self, secret: Secret[str]):
        with pytest.raises(SecretException):
            hashlib.sha1(secret.encode())

    def test_secret_sha1_succeeds(self, secret: Secret[str]):
        sha1 = secret.cast(bytes).dangerous_map(lambda x: hashlib.sha1(x).hexdigest())

        with pytest.raises(SecretException):
            assert sha1 == b"6ffd8b80f2a76ca670ae33ab196f7936d59fb43b"

        with sha1.cast(bytes).dangerous_reveal() as revealed:
            assert revealed == b"6ffd8b80f2a76ca670ae33ab196f7936d59fb43b"

    def test_secret_kdf(self, secret: Secret[str]):
        salt = os.urandom(16)
        kdf = Scrypt(salt=salt, length=32, n=2**14, r=8, p=1)

        with pytest.raises(SecretException):
            kdf.derive(secret.encode())

        key = secret.cast(bytes).dangerous_map(kdf.derive)

        # store key as plaintext
        key_plain = key._dangerous_extract()
        with pytest.raises(SecretException):
            kdf.verify(secret.encode(), key_plain)

        # verify key
        kdf = Scrypt(salt=salt, length=32, n=2**14, r=8, p=1)
        secret.cast(bytes).dangerous_apply(kdf.verify, key_plain)

    def test_secret_concat(self, secret: Secret[str]):
        result = secret + "foobar"

        with pytest.raises(SecretException):
            print(result)

        with result.dangerous_reveal() as revealed:
            assert revealed == "foobar123foobar"

    def test_secret_multiply(self, secret: Secret[str]):
        result = 1 * secret * 3

        with pytest.raises(SecretException):
            print(result)

        with result.dangerous_reveal() as revealed:
            assert revealed == "foobar123" * 3

    def test_int_secret(self, int_secret: SecretNumber[int]):
        one_hundred = int_secret + 58
        one_thousand_six_hundred = one_hundred << 4
        one_hundred_squared = (one_thousand_six_hundred >> 4) * one_hundred

        one_hundred_again = one_hundred_squared.dangerous_map(math.sqrt)
        two_hundred = one_hundred_again + one_hundred

        with pytest.raises(SecretException):
            print(two_hundred)

        with two_hundred.cast(int).dangerous_reveal() as revealed:
            assert revealed is 200

    def test_token(self):
        token = Secret.token(32)

        with pytest.raises(SecretException):
            print(token)

        with token.dangerous_reveal() as revealed:
            assert len(revealed) == 32



from typing import TYPE_CHECKING, Optional, Tuple, Union

if TYPE_CHECKING:
    from secret_type.containers.number import SecretNumber


class IntegerOps:
    BI_OPS = [
        "add",
        "sub",
        "mul",
        "truediv",
        "floordiv",
        "mod",
        "divmod",
        "pow",
        "lshift",
        "rshift",
        "and",
        "xor",
        "or",
    ]
    UNI_OPS = [
        "neg",
        "pos",
        "abs",
        "invert",
        "round",
        "trunc",
        "floor",
        "ceil",
        "lt",
        "le",
    ]

    def __add__(self, __x: Union[int, "SecretNumber"]) -> "SecretNumber[int]":
        ...

    def __sub__(self, __x: Union[int, "SecretNumber"]) -> "SecretNumber[int]":
        ...

    def __mul__(self, __x: Union[int, "SecretNumber"]) -> "SecretNumber[int]":
        ...

    def __floordiv__(self, __x: Union[int, "SecretNumber"]) -> "SecretNumber[int]":
        ...

    def __truediv__(self, __x: Union[int, "SecretNumber"]) -> float:
        ...

    def __mod__(self, __x: Union[int, "SecretNumber"]) -> "SecretNumber[int]":
        ...

    def __divmod__(self, __x: Union[int, "SecretNumber"]) -> Tuple[int, int]:
        ...

    def __radd__(self, __x: Union[int, "SecretNumber"]) -> "SecretNumber[int]":
        ...

    def __rsub__(self, __x: Union[int, "SecretNumber"]) -> "SecretNumber[int]":
        ...

    def __rmul__(self, __x: Union[int, "SecretNumber"]) -> "SecretNumber[int]":
        ...

    def __rfloordiv__(self, __x: Union[int, "SecretNumber"]) -> "SecretNumber[int]":
        ...

    def __rtruediv__(self, __x: Union[int, "SecretNumber"]) -> float:
        ...

    def __rmod__(self, __x: Union[int, "SecretNumber"]) -> "SecretNumber[int]":
        ...

    def __rdivmod__(self, __x: Union[int, "SecretNumber"]) -> Tuple[int, int]:
        ...

    def __pow__(
        self, __x: Union[int, "SecretNumber"], __modulo: int
    ) -> "SecretNumber[int]":
        ...

    def __rpow__(
        self, __x: Union[int, "SecretNumber"], __mod: Optional[int] = ...
    ) -> "SecretNumber[int]":
        ...

    def __and__(self, __n: Union[int, "SecretNumber"]) -> "SecretNumber[int]":
        ...

    def __or__(self, __n: Union[int, "SecretNumber"]) -> "SecretNumber[int]":
        ...

    def __xor__(self, __n: Union[int, "SecretNumber"]) -> "SecretNumber[int]":
        ...

    def __lshift__(self, __n: Union[int, "SecretNumber"]) -> "SecretNumber[int]":
        ...

    def __rshift__(self, __n: Union[int, "SecretNumber"]) -> "SecretNumber[int]":
        ...

    def __rand__(self, __n: Union[int, "SecretNumber"]) -> "SecretNumber[int]":
        ...

    def __ror__(self, __n: Union[int, "SecretNumber"]) -> "SecretNumber[int]":
        ...

    def __rxor__(self, __n: Union[int, "SecretNumber"]) -> "SecretNumber[int]":
        ...

    def __rlshift__(self, __n: Union[int, "SecretNumber"]) -> "SecretNumber[int]":
        ...

    def __rrshift__(self, __n: Union[int, "SecretNumber"]) -> "SecretNumber[int]":
        ...

    def __neg__(self) -> "SecretNumber[int]":
        ...

    def __pos__(self) -> "SecretNumber[int]":
        ...

    def __invert__(self) -> "SecretNumber[int]":
        ...

    def __trunc__(self) -> "SecretNumber[int]":
        ...

    def __ceil__(self) -> "SecretNumber[int]":
        ...

    def __floor__(self) -> "SecretNumber[int]":
        ...

    def __round__(self, __ndigits: int = ...) -> "SecretNumber[int]":
        ...

    def __getnewargs__(self) -> Tuple[int]:
        ...

    def __eq__(self, __x: object) -> bool:
        ...

    def __ne__(self, __x: object) -> bool:
        ...

    def __lt__(self, __x: Union[int, "SecretNumber"]) -> bool:
        ...

    def __le__(self, __x: Union[int, "SecretNumber"]) -> bool:
        ...

    def __gt__(self, __x: Union[int, "SecretNumber"]) -> bool:
        ...

    def __ge__(self, __x: Union[int, "SecretNumber"]) -> bool:
        ...

    def __abs__(self) -> "SecretNumber[int]":
        ...



"""This module contains helper types that are used by the rest of the library."""

from numbers import Integral, Number, Rational
from typing import TypeVar, Union

from typing_extensions import ParamSpec

StringLike = Union[str, bytes]
"""A type that can be treated as a string.
[`Secret[StringLike]`][secret_type.Secret] support all `str` operations."""
IntLike = Union[int, Integral, float, Rational]
"""A numeric type that supports integer operations.

[`Secret[IntLike]`][secret_type.Secret] support all `int` operations."""
NumberLike = Union[complex, Number]
"""A numeric type that does not support integer operations.

[`Secret[NumberLike]`][secret_type.Secret] support only a limited set of numeric operations."""
BoolLike = TypeVar("BoolLike", bound=bool)
"""A boolean type."""
ProtectedValue = Union[StringLike, IntLike, NumberLike, bool]
"""A type that can be wrapped in a [`Secret`][secret_type.Secret]."""


T = TypeVar("T", bound=ProtectedValue)
"""A [`ProtectedValue`][secret_type.typing.types.ProtectedValue]."""
T2 = TypeVar("T2", bound=ProtectedValue)
"""Another [`ProtectedValue`][secret_type.typing.types.ProtectedValue]."""
T_con = TypeVar("T_con", bound=ProtectedValue, contravariant=True)
T_co = TypeVar("T_co", bound=ProtectedValue, covariant=True)

S = TypeVar("S", bound=StringLike)
"""A [`StringLike`][secret_type.typing.types.StringLike]."""
N = TypeVar("N", bound=IntLike)
"""An [`IntLike`][secret_type.typing.types.IntLike]."""

R = TypeVar("R")
P = ParamSpec("P")



"""This module contains exceptions that are used by the rest of the library."""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from secret_type.containers.secret import Secret


class SecretException(Exception):
    """The base exception for this library.

    All other exceptions are subclasses of this one."""

    def __init__(self, message: str = "Secrets cannot be examined") -> None:
        super().__init__(message)


class SecretFloatException(SecretException):
    """Raised when an unsupported operation is attempted on a [`Secret[NumberLike]`][secret_type.typing.types.NumberLike]."""

    def __init__(
        self,
        message: str = "Secrets cannot be used as non-integral numbers",
    ) -> None:
        super().__init__(message)


class SecretKeyException(SecretException):
    """Raised when a [`Secret`][secret_type.Secret] is used as a key or index."""

    def __init__(
        self,
        message: str = "Secrets cannot be used as keys",
    ) -> None:
        super().__init__(message)


class SecretBoolException(SecretException):
    """Raised when a [`Secret[bool]`][secret_type.Secret] is used for control flow."""

    def __init__(
        self,
        message: str = "bools derived from Secrets cannot be used for control flow",
    ) -> None:
        super().__init__(message)


class SecretAttributeError(AttributeError, SecretException):
    def __init__(self, s: "Secret", name: str) -> None:
        message = f"{s.protected_type.__name__} has no attribute {name}"
        super().__init__(message)



# SPDX-FileCopyrightText: 2022-present Yasyf Mohamedali <yasyfm@gmail.com>
#
# SPDX-License-Identifier: MIT
"""`secret-type` provides a convenient type (`secret`) to indicate that a value is considered sensitive,
    similar to the `secret` type in Google's [Rune Lang](https://github.com/google/rune).

    ```pycon
    >>> from secret_type import secret
    >>> password = secret("a very secret value") # (1)!

    >>> print(password) # (2)!
    Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
      File "secret_type/containers/secret.py", line 91, in __str__
        raise SecretException()
    secret_type.exceptions.SecretException: Secrets cannot be examined

    >>> better_password = password + "!" # (3)!
    >>> >>> type(better_password)
    <class 'secret_type.sequence.SecretStr'>

    >>> better_password.dangerous_apply(print) # (4)!
    a very secret value!
    ```

    1. Secrets can be any primitive value
    2. Runtime exceptions prevent logging
    3. Operations derive new secrets
    4. Use [`dangerous_apply`][secret_type.Secret.dangerous_apply] or [`dangerous_map`][secret_type.Secret.dangerous_map] to access the underlying value

    # Docs

    For complete docs, see the [Quickstart](quickstart.md), or check out the [API Reference](reference/index.md).

    # Features
      - When marked as secret, values cannot be printed or logged; attempting to do so will raise an exception.
      - Secrets are "viral"; any operation on a secret will also return a secret.
      - Comparison operations with a `secret` are guaranteed to be constant-time.
        This helps avoid timing attacks.
      - A `bool` derived from a secret cannot be used for control flow.
      - Secrets cannot be used as indexes or keys for containers.
      - Internally, the underlying value is stored encrypted in memory, and is only decrypted when deriving a new value.
      - As soon as secrets are out of scope, the Garbage Collector is encouraged to immediately collect them.

    # Comparison to Rune
    Rune makes the following guarantees about a `secret`:

    > - All operations on secrets occur in constant time, minimizing timing side-channel leakage.
    - Secrets cannot be used in conditional branches or memory addressing.
    - Even speculative branching and indexing on secrets are caught at compile-time to avoid Specter/Meltdown.
    - Secrecy is sticky: any value in part derived from a secret is considered secret until "revealed".
    - Secrets are automatically zeroed when no longer used

    This projects attempts to do something similar, but with the runtime constraints of Python.
    """

from typing import Union as Union

from secret_type.containers.secret import Secret as Secret
from secret_type.monad import SecretMonad as SecretMonad
from secret_type.typing.types import T


def secret(o: T) -> Secret[T]:
    """This single function provides a convenient to indicate that a value is considered sensitive.

    Sensitive values can be any (subclass of a) non-container primitive.
    Simply wrap the value in a call to `secret`, and off you go!
    Any operations you do on that wrapped value will also return a `secret`.

    Attributes:
      o (Union[str, bytes, int, float, bool]): The value to mark as secret.

    Examples: Example:
      Here is an example using `secret` to protect a key derived from a user password.
      We want to make sure that no one is able to accidentally log either of these values,
      that all comparison operations are constant-time, and that any derived values are also secret.

      ```python
      from secret_type import secret
      from cryptography.hazmat.primitives.kdf.scrypt import Scrypt # (1)!

      salt = os.urandom(16) # (2)!

      def derive_key(user_password: str):
        password = secret(user_password)
        kdf = Scrypt(salt=salt, length=32, n=2**14, r=8, p=1)

        key = password.cast(bytes).dangerous_map(kdf.derive) # (3)!

        with key.dangerous_reveal() as key: # (4)!
          persist_to_database(key)

      # Some time later...

      def check_password(entered_password: str):
        password = secret(entered_password)
        kdf = Scrypt(salt=salt, length=32, n=2**14, r=8, p=1)

        key = secret(get_key_from_database()) # (5)!
        password.cast(bytes).dangerous_apply(kdf.verify, key) # (6)!
      ```

      1. We're using `Scrypt` as an example, but `secret` is not specific to the `cryptography` package.
      2. A random salt that does not need to be secret.
      3. We first use [`cast`][secret_type.Secret.cast] to convert the `secret` to a `bytes` object,
        then use [`dangerous_map`][secret_type.Secret.dangerous_map] to run the key derivation function,
        wrapping the result in a new `secret`.
      4. We must use [`dangerous_reveal`][secret_type.Secret.dangerous_reveal] to expose the actual key,
        which we persist to a database.
      5. We retreive the plain-text key from the database, and immediately call `secret` on it to resume
        our protections.
      6. Here we use [`dangerous_apply`][secret_type.Secret.dangerous_apply] to run the key verification function,
        which raises if it fails. We use this instead of [`dangerous_map`][secret_type.Secret.dangerous_map] as we don't care about the return value.
    """
    return Secret.wrap(o)



from numbers import Number, Rational
from typing import TYPE_CHECKING, Union, overload

if TYPE_CHECKING:
    from secret_type.containers.secret import Secret

from secret_type.typing.types import R, T


class SecretMonad:
    """Provides a monad-like interface for [`Secret`][secret_type.Secret] objects.

    These methods are all also available on the [`Secret`][secret_type.Secret] class.
    """

    @overload
    @classmethod
    def wrap(cls, o: "Secret[T]") -> "Secret[T]":
        ...

    @overload
    @classmethod
    def wrap(cls, o: T) -> "Secret[T]":
        ...

    @classmethod
    def wrap(cls, o: Union[T, "Secret[T]"]) -> "Secret[T]":
        """Wraps a value in the appropriate [`Secret`][secret_type.Secret] container.

        If the value is already a [`Secret`][secret_type.Secret], it is returned as-is.

        Attributes:
            o (Union[str, bytes, int, float, bool]): The value to wrap.

        Examples: Example:
            ```python
            secret = SecretMonad.wrap(42)
            ```

        Raises:
            TypeError: If `o` is not a primitive value.
        """
        from secret_type.containers.bool import SecretBool
        from secret_type.containers.number import SecretNumber
        from secret_type.containers.secret import Secret
        from secret_type.containers.sequence import SecretStr

        if isinstance(o, Secret):
            return o
        elif isinstance(o, (str, bytes)):
            return SecretStr(o)
        elif isinstance(o, bool):
            return SecretBool(o)
        elif isinstance(o, Rational):
            return SecretNumber(o)
        elif isinstance(o, Number):
            return Secret(o)
        else:
            raise TypeError("Cannot wrap type '{}'".format(type(o).__name__))

    @overload
    @classmethod
    def unwrap(cls, o: "Secret[T]") -> T:
        ...

    @overload
    @classmethod
    def unwrap(cls, o: R) -> R:
        ...

    @classmethod
    def unwrap(cls, o: Union["Secret[T]", R]) -> Union[T, R]:
        """Unwraps a [`Secret`][secret_type.Secret] into it's underlying value.

        If the value is not a [`Secret`][secret_type.Secret], it is returned as-is.

        Attributes:
            o (Secret[T]): The value to unwrap.

        Examples: Example:
            ```python
            assert 42 == SecretMonad.unwrap(secret)
            ```
        """
        from secret_type.containers.secret import Secret

        if isinstance(o, Secret):
            return o._dangerous_extract()
        else:
            return o



"""This class contains specialized containers for holding secrets of various types."""

from secret_type.containers.secret import (  # noqa # isort:skip
    Secret as Secret,
)

from secret_type.containers.bool import SecretBool as SecretBool
from secret_type.containers.number import SecretNumber as SecretNumber
from secret_type.containers.sequence import SecretStr as SecretStr



from secret_type.containers.secret import Secret
from secret_type.exceptions import SecretBoolException
from secret_type.typing.types import BoolLike


class SecretBool(Secret[BoolLike]):
    """A specialized subclass of [`Secret[bool]`][secret_type.Secret] for holding bool results.

    This class is returned whenever a `bool` is derived from an operation on a [`Secret`][secret_type.Secret].
    It ensures that the result cannot be used for control flow,
    unless explicitly allowed by using [`dangerous_reveal`][secret_type.Secret.dangerous_reveal].
    """

    def flip(self):
        """Flip the value of the contained bool without revealing it."""
        return SecretBool(not self._dangerous_extract())

    def __bool__(self):
        raise SecretBoolException()

    def __eq__(self, other):
        return self.dangerous_map(lambda x: x == other)

    def __repr__(self):
        return repr(self._dangerous_extract())

    def __str__(self):
        return str(self._dangerous_extract())
